<?php
declare(strict_types=1);

namespace IMATHUZH\Qfq\Core\Form;

use IMATHUZH\Qfq\Core\Database\Database;
use IMATHUZH\Qfq\Core\Helper\HelperFile;
use IMATHUZH\Qfq\Core\Helper\OnString;
use IMATHUZH\Qfq\Core\Helper\Path;
use IMATHUZH\Qfq\Core\Helper\SqlQuery;
use IMATHUZH\Qfq\Core\Store\Store;

class FormAsFile
{
    /**
     * Remove the form from the DB and insert it using the form file. (only if the form file was changed)
     * If the form file can't be read, then the form is deleted from the DB and an exception is thrown.
     *
     * Form file location: SYSTEM_FORM_FILE_PATH
     * Recognize form file change: Compare the current file stats with the ones saved in the Form table.
     * Container References: The form file uses names instead if ids to reference container formElements. These references are translated when the formElements are inserted.
     * Ignored keys: The keys 'id', 'formId', 'feIdContainer' are ignored when reading the form file.
     *
     * @param string $formName
     * @param Database $database
     * @return bool True if form has been updated.
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    public static function importForm(string $formName, Database $database): bool
    {
        // Get file stats both from file system and DB and exit if they are equal
        $pathFileName = self::formPathFileName($formName, $database);
        $fileReadException = new \UserFormException(json_encode([
            ERROR_MESSAGE_TO_USER => "Form file not found or missing permission: "  . baseName($pathFileName),
            ERROR_MESSAGE_TO_DEVELOPER => "Form definition file not found or no permission to read file: '$pathFileName'"]),
            ERROR_FORM_NOT_FOUND);
        if(!file_exists($pathFileName)) {
            self::deleteFormDB($formName, $database);
            throw $fileReadException;
        }
        $fileStatsNew = self::formFileStatsJson($pathFileName);
        if ($fileStatsNew === false) {
            self::deleteFormDB($formName, $database);
            throw $fileReadException;
        }
        list($sql, $parameterArray) = SqlQuery::selectFormByName($formName, [F_ID, F_FILE_STATS]);
        $formFromDb = $database->sql($sql, ROW_EXPECT_0_1,
            $parameterArray, "Multiple forms with the same name: '$formName'");
        if (array_key_exists(F_FILE_STATS, $formFromDb) && $fileStatsNew === $formFromDb[F_FILE_STATS]) {
            return false;
        }

        // Read form file
        $fileContents = file_get_contents($pathFileName);
        if ($fileContents === false) {
            self::deleteFormDB($formName, $database);
            throw $fileReadException;
        }
        $formFromFile = json_decode($fileContents, true);

        // Delete old form with that name from DB if it exists.
        if (array_key_exists(F_ID, $formFromDb)) {
            $formId = $formFromDb[F_ID];
            self::deleteFormDBWithId($formId, $database);
        }

        // Insert new Form to DB (after filtering allowed columns and adding column 'name')
        $formSchema = $database->getTableDefinition(TABLE_NAME_FORM);
        $formColumns = array_column($formSchema, 'Field');
        $insertValues = array_filter($formFromFile, function ($columnName) use ($formColumns) {
            return $columnName !== F_ID && in_array($columnName, $formColumns);
        }, ARRAY_FILTER_USE_KEY); // array(column => value)
        $insertValues[F_NAME] = $formName;
        list($sqlFormInsert, $parameterArrayFormInsert) = SqlQuery::insertRecord(TABLE_NAME_FORM, $insertValues);
        $formId = $database->sql($sqlFormInsert, ROW_REGULAR, $parameterArrayFormInsert);

        // Delete stale formElements with the new form id (these should not exist, but better make sure)
        self::deleteFormElementsDBWithFormId($formId, $database);

        // Insert FormElements to DB and collect container ids
        $containerIds = []; // array(container_name => id)
        foreach ($formFromFile[F_FILE_FORM_ELEMENT] as &$formElementFromFile) {
            $feId = self::insertFormElement($formElementFromFile, $formId, $database);
            $formElementFromFile[FE_ID] = $feId;
            if ($formElementFromFile[FE_CLASS] === FE_CLASS_CONTAINER) {
                $containerIds[$formElementFromFile[FE_NAME]] =  $feId;
            }
        }

        // Update container IDs for each form element which has a container name
        foreach ($formFromFile[F_FILE_FORM_ELEMENT] as &$formElementFromFile) {
            if (array_key_exists(FE_FILE_CONTAINER_NAME, $formElementFromFile)) {
                $containerName = $formElementFromFile[FE_FILE_CONTAINER_NAME];
                $containerId = $containerIds[$containerName];
                $feId = $formElementFromFile[FE_ID];
                list($sql, $parameterArray) = SqlQuery::updateRecord(TABLE_NAME_FORM_ELEMENT, [FE_ID_CONTAINER => $containerId], $feId);
                $database->sql($sql, ROW_REGULAR, $parameterArray);
            }
        }

        // Update column fileStats if everything went well
        list($sql, $parameterArray) = SqlQuery::updateRecord(TABLE_NAME_FORM, [F_FILE_STATS => $fileStatsNew], $formId);
        $database->sql($sql, ROW_REGULAR, $parameterArray);

        return true;
    }

    /**
     * Reads the form from the DB and saves it in the form folder as a form file.
     * Warning: Overwrites form file without any checks or warning.
     * If the form file path does not exist, it is created and all forms are exported.
     *
     * FileStats: After the export the column "fileStats" is updated with new stats from new file.
     * Container FormElements: The form file uses names instead if ids to reference containers. These references are translated before saving the file.
     * Ignored columns: The columns 'id', 'name' (only Form table), 'fileStats', 'formId', 'feIdContainer' are not saved in the file.
     * FormElement order: The formElements are ordered by the 'ord' column before writing them to the file.
     *
     * @param string $formName
     * @param Database $database
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    public static function exportForm(string $formName, Database $database) // : void
    {
        // Get form from DB
        list($sql, $parameterArray) = SqlQuery::selectFormByName($formName);
        $form = $database->sql($sql, ROW_EXPECT_1,
            $parameterArray, 'Form "' . $formName . '" not found or multiple forms with the same name.'); // array(column name => value)

        // Remove columns: id, name, fileStats
        $formId = $form[F_ID];
        unset($form[F_ID]);
        unset($form[F_NAME]);
        unset($form[F_FILE_STATS]);

        // Get formElements from DB
        list($sql, $parameterArray) = SqlQuery::selectFormElementById($formId);
        $formElements = $database->sql($sql, ROW_REGULAR, $parameterArray); // array(array(column name => value))

        // Translate container references (id to name) and remove all id columns
        $containerNames = array_reduce($formElements, function ($result, $formElement) {
            if ($formElement[FE_CLASS] === FE_CLASS_CONTAINER) {
                $containerName = $formElement[FE_NAME];
                if (in_array($containerName, $result) || $containerName === '') {
                    throw new \UserFormException(json_encode([
                        ERROR_MESSAGE_TO_USER => 'Duplicate container names.',
                        ERROR_MESSAGE_TO_DEVELOPER => "Container Form Elements must have a unique and nonempty name. Container name: '$containerName'."]),
                        ERROR_FORM_INVALID_NAME);
                }
                $result[$formElement[FE_ID]] = $containerName;
            }
            return $result;
        }, []); // array(id => name)
        $formElements = array_map(function ($formElement) use ($containerNames) {
            $containerId = $formElement[FE_ID_CONTAINER];
            if ($containerId !== 0) {
                $formElement[FE_FILE_CONTAINER_NAME] = $containerNames[$containerId];
            }
            unset($formElement[FE_ID_CONTAINER]);
            unset($formElement[FE_ID]);
            unset($formElement[FE_FORM_ID]);
            return $formElement;
        }, $formElements);

        // write form as JSON to file
        $form[F_FILE_FORM_ELEMENT] = $formElements;
        $formJson = json_encode($form, JSON_PRETTY_PRINT);
        $pathFileName = self::formPathFileName($formName, $database);
        HelperFile::file_put_contents($pathFileName, $formJson);

        // Update column fileStats
        $fileStats = self::formFileStatsJson($pathFileName);
        list($sql, $parameterArray) = SqlQuery::updateRecord(TABLE_NAME_FORM, [F_FILE_STATS => $fileStats], $formId);
        $database->sql($sql, ROW_REGULAR, $parameterArray);
    }

    /**
     * Import Form from file if loaded record is Form/FormElement.
     * If form file was changed, import it and throw exception.
     * If form file is not writeable, throw exception.
     *
     * @param int $recordId
     * @param string $tableName
     * @param Database $database
     * @return string|null formName|null
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    public static function importFormRecordId(int $recordId, string $tableName, Database $database) // : ?string
    {
        $recordFormName = self::formNameFromFormRelatedRecord($recordId, $tableName, $database);
        if ($recordFormName !== null) {
            if(self::importForm($recordFormName, $database)) {
                throw new \UserFormException(json_encode([
                    ERROR_MESSAGE_TO_USER => 'Form file was changed.',
                    ERROR_MESSAGE_TO_DEVELOPER => "Form definition file has been changed. Please close tab and reload the form list and Form-Editor from scratch."]),
                    ERROR_FORM_NOT_FOUND);
            }
            self::enforceFormFileWritable($recordFormName, $database);
        }
        return $recordFormName;
    }

    /**
     * Deletes the form file for the given form.
     *
     * @param $formName
     * @param Database $database
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    public static function deleteFormFile($formName, Database $database) // : void
    {
        self::enforceFormFileWritable($formName, $database);
        $pathFileName = self::formPathFileName($formName, $database);
        if(file_exists($pathFileName)) {
            $success = unlink($pathFileName);
            if ($success === false) {
                throw new \UserFormException(json_encode([
                    ERROR_MESSAGE_TO_USER => "Deleting form file failed: " . baseName($pathFileName),
                    ERROR_MESSAGE_TO_DEVELOPER => "Can't delete form file '$pathFileName'"]),
                    ERROR_IO_WRITE_FILE);
            }
        }
    }

    /**
     * Multiple errors might occur after a form file import. This hint can be added to such exceptions to help users.
     * Only returns non-empty string if the table is either Form of FormElement.
     *
     * @param string $tableName
     * @return string
     */
    public static function errorHintFormImport (string $tableName = TABLE_NAME_FORM): string
    {
        $message = '';
        if (in_array($tableName, [TABLE_NAME_FORM, TABLE_NAME_FORM_ELEMENT])) {
            $message .= "Hint: Form definition file might have changed. Please reopen the form list and Form-Editor from scratch.";
        }
        return $message;
    }

    /**
     * Throw exception if form file or form directory is not writeable.
     *
     * @param string $formName
     * @param Database $database
     * @return void
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    public static function enforceFormFileWritable(string $formName, Database $database) // : void
    {
        $pathFileName = self::formPathFileName($formName, $database);
        HelperFile::enforce_writable_or_creatable($pathFileName);

    }

    /**
     * Return form name if given record is an existing Form or FormElement.
     * Return null otherwise.
     *
     * @param int $recordId
     * @param string $recordTable
     * @param Database $database
     * @return string formName|null
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    public static function formNameFromFormRelatedRecord(int $recordId, string $recordTable, Database $database) // : ?string
    {
        if ($recordId === 0) {
            return null;
        }
        switch ($recordTable) {
            case TABLE_NAME_FORM:
                list($sql, $parameterArray) = SqlQuery::selectFormById($recordId);
                $formFromDb = $database->sql($sql, ROW_EXPECT_1,
                    $parameterArray, "Form with id '$recordId' not found. " . self::errorHintFormImport());
                return $formFromDb[F_NAME];
            case TABLE_NAME_FORM_ELEMENT:
                $F_NAME = F_NAME;
                $TABLE_NAME_FORM_ELEMENT = TABLE_NAME_FORM_ELEMENT;
                $TABLE_NAME_FORM = TABLE_NAME_FORM;
                $F_ID = F_ID;
                $FE_FORM_ID = FE_FORM_ID;
                $FE_ID = FE_ID;

                // Select form name by formElement id
                $formFromDb = $database->sql("SELECT `f`.`$F_NAME` FROM `$TABLE_NAME_FORM` AS f INNER JOIN `$TABLE_NAME_FORM_ELEMENT` AS fe ON f.`$F_ID`=fe.`$FE_FORM_ID` WHERE `fe`.`$FE_ID`=?", ROW_EXPECT_1,
                    [$recordId], "Form element with id '$recordId' not found. " . self::errorHintFormImport());
                return $formFromDb[F_NAME];
            default:
                return null;
        }
    }

    /**
     * Returns true if the sql query selects Form or FormElement table
     *
     * @param string $sql
     * @return bool
     */
    public static function isFormQuery(string $sql): bool
    {
        // find substrings whihch start with FROM and are followed by Form or FormElement
        preg_match_all('/(?i)FROM(?-i)(.*?)\b(' . TABLE_NAME_FORM . '|' . TABLE_NAME_FORM_ELEMENT . ')\b/s', $sql, $matches);

        // Check if no other SQL keywords are in between FROM and the table name
        $keywordsAfterFrom = ['WHERE', 'GROUP BY', 'HAVING', 'WINDOW', 'ORDER BY', 'LIMIT', 'FOR', 'INTO'];
        foreach($matches[0] as $match)
        {
            if(!OnString::containsOneOfWords($keywordsAfterFrom, $match)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Import all form files into the database.
     * If the file folder does not exist, it is created and all forms from the DB are exported.
     *
     * @param Database $database
     * @param bool $enforceWritable Throw exception if one of the form files is not writable by current user.
     * @param bool $deleteFromDB Delete forms from DB if there are no corresponding form files.
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    public static function importAllForms(Database $database, bool $enforceWritable = false, bool $deleteFromDB = false) // : void
    {
        // Import all form files
        $formFileNames = self::formFileNames($database);
        foreach ($formFileNames as $formFileName) {
            self::importForm($formFileName, $database);
            if ($enforceWritable) {
                self::enforceFormFileWritable($formFileName, $database);
            }
        }

        // Delete all forms which are in DB but not in files
        if ($deleteFromDB) {
            $formNamesDB = self::queryAllFormNames($database);
            $formsToDelete = array_diff($formNamesDB, $formFileNames);
            foreach ($formsToDelete as $formToDelete) {
                self::deleteFormDB($formToDelete, $database);
            }
        }
    }

    /**
     * Export all forms in database to form files.
     * Warning: This overwrites form files without any checks or warning. Use cautiously.
     *
     * @param Database $database
     * @param bool $deleteFiles Delete form files without a corresponding form record in the DB.
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    public static function exportAllForms(Database $database, bool $deleteFiles = false) // : void
    {
        $formNamesDB = self::queryAllFormNames($database);
        foreach ($formNamesDB as $formNameDB) {
            self::exportForm($formNameDB, $database);
        }
        if ($deleteFiles) {
           $formFileNames = self::formFileNames($database);
           $filesToDelete = array_diff($formFileNames, $formNamesDB);
           foreach ($filesToDelete as $fileToDelete) {
               self::deleteFormFile($fileToDelete, $database);
           }
        }
    }

    /**
     * Insert formElement to the given form.
     * Keys removed before insert: id, formId, feIdContainer
     *
     * @param array $values
     * @param int $formId
     * @param Database $database
     * @return int
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    private static function insertFormElement(array $values, int $formId, Database $database): int
    {
        // filter allowed formElement columns
        $formElementSchema = $database->getTableDefinition(TABLE_NAME_FORM_ELEMENT);
        $formElementColumns = array_column($formElementSchema, 'Field');
        $insertValues = array_filter($values, function ($columnName) use ($formElementColumns) {
            return !in_array($columnName, [FE_ID, FE_FORM_ID, FE_ID_CONTAINER]) && in_array($columnName, $formElementColumns);
        }, ARRAY_FILTER_USE_KEY); // array(column => value)

        // execute insert
        $insertValues[FE_FORM_ID] = $formId;
        list($sql, $parameterArray) = SqlQuery::insertRecord(TABLE_NAME_FORM_ELEMENT, $insertValues);
        $id = $database->sql($sql, ROW_REGULAR, $parameterArray);

        return $id;
    }

    /**
     * Delete form with given name and its form elements from DB
     *
     * @param string $formName
     * @param Database $database
     * @return void
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    private static function deleteFormDB(string $formName, Database $database) // : void
    {
        list($sql, $parameterArray) = SqlQuery::selectFormByName($formName, [F_ID]);
        $formFromDb = $database->sql($sql, ROW_EXPECT_0_1,
            $parameterArray, "Multiple forms with the same name: '$formName'");
        if (array_key_exists(F_ID, $formFromDb)) {
            self::deleteFormDBWithId($formFromDb[F_ID], $database);
        }
    }

    /**
     * Delete form with given id and its form elements from DB
     *
     * @param $formId
     * @param Database $database
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    private static function deleteFormDBWithId(int $formId, Database $database) // : void
    {
        $F_ID = F_ID; // can't use constants in strings directly
        $TABLE_NAME_FORM = TABLE_NAME_FORM; // can't use constants in strings directly
        $database->sql("DELETE FROM `$TABLE_NAME_FORM` WHERE `$F_ID`=? LIMIT 1", ROW_REGULAR, [$formId]);
        self::deleteFormElementsDBWithFormId($formId, $database);
    }

    /**
     * Delete form elements with given formId from DB
     *
     * @param $formId
     * @param Database $database
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    private static function deleteFormElementsDBWithFormId(int $formId, Database $database) // : void
    {
        $TABLE_NAME_FORM_ELEMENT = TABLE_NAME_FORM_ELEMENT; // can't use constants in strings directly
        $FE_FORM_ID = FE_FORM_ID; // can't use constants in strings directly
        $database->sql("DELETE FROM `$TABLE_NAME_FORM_ELEMENT` WHERE `$FE_FORM_ID`=?", ROW_REGULAR, [$formId]);
    }

    /**
     * Return a selection of file stats as a JSON string which serves as a fingerprint of the file.
     * If any one of the stats changed the file content was probably changed as well.
     *
     * @param string $pathFileName
     * @return false|string
     */
    private static function formFileStatsJson(string $pathFileName)
    {
        $stats = stat($pathFileName);
        if ($stats === false) {
            return false;
        }
        return json_encode([
            'modified' => $stats['mtime'],
            'size' => $stats['size'],
            'inode' => $stats['ino']
        ]);
    }

    /**
     * Return correct pathFileName of form file relative to current working directory.
     * Create path if it doesn't exist and export all forms from DB.
     *
     * @param string $formName
     * @param Database $database
     * @return string
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    private static function formPathFileName(string $formName, Database $database): string
    {
        // validate form name
        if (!HelperFile::isValidFileName($formName)) {
            throw new \UserFormException(json_encode([
                ERROR_MESSAGE_TO_USER => 'Reading/Writing Form file failed.',
                ERROR_MESSAGE_TO_DEVELOPER => "Form name '$formName' not valid. Name may only consist of alphanumeric characters and _ . -"]),
                ERROR_FORM_INVALID_NAME);
        }
        return self::formPath($database) . '/' . $formName . ".json";
    }

    /**
     * Return the path of the form directory relative to CWD.
     * Create path if it doesn't exist and export all forms from DB.
     *
     * @param Database $database
     * @return string
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    private static function formPath(Database $database): string
    {
        $cwdToForm = Path::cwdToProject(Path::PROJECT_TO_FORM);
        if (!is_dir($cwdToForm)) {

            // create path
            $success = mkdir($cwdToForm, 0777, true);
            if ($success === false) {
                throw new \UserFormException(json_encode([
                    ERROR_MESSAGE_TO_USER => "Can't create form file path.",
                    ERROR_MESSAGE_TO_DEVELOPER => "Can't create path: " . $cwdToForm]),
                    ERROR_IO_WRITE_FILE);
            }

            // export all forms
            self::exportAllForms($database);
        }
        return $cwdToForm;
    }

    /**
     * Return array of all form names in the Form table.
     *
     * @param Database $database
     * @return array [formName]
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    private static function queryAllFormNames(Database $database): array
    {
        $NAME = F_NAME;
        $FORM = TABLE_NAME_FORM;
        return array_column($database->sql("SELECT `$NAME` FROM `$FORM`", ROW_REGULAR), $NAME);
    }

    /**
     * Return array of form file names contained in the form path (without suffix).
     *
     * @param Database $database
     * @return array|false [formName]
     * @throws \CodeException
     * @throws \DbException
     * @throws \UserFormException
     */
    private static function formFileNames(Database $database): array
    {
        $formPath = self::formPath($database);
        $files = scandir($formPath);
        if ($files === false) {
            throw new \UserFormException(json_encode([
                ERROR_MESSAGE_TO_USER => "Reading directory failed.",
                ERROR_MESSAGE_TO_DEVELOPER => "Can't read directory: " . $formPath]),
                ERROR_IO_READ_FILE);
        }
        return $jsonFileNames = array_reduce($files, function ($result, $file) {
            $fileInfo = pathinfo($file);
            if ($fileInfo['extension'] === 'json') {
                $result[] = $fileInfo['filename'];
            }
            return $result;
        }, []);
    }
}