<?php


namespace IMATHUZH\Qfq\Core\Helper;

/*
 * Naming convention of path constants/functions/variables:
 * 1) name a path by its origin and its destination separated by 'to'. E.g. APP_TO_SYSTEM_LOG, $appToProject.
 * 2) if the destination is a file, append "File". E.g. APP_TO_SYSTEM_QFQ_LOG_FILE.
 * 3) if a path has to be variable, create a setter and getter. E.g. self::setCwdToApp(), self::cwdToApp(), private static $cwdToApp.
 * 4) a path getter appends the given arguments to the requested path using self::join(..., func_get_args()). E.g. see cwdToApp().
 * 5) additional path getters may be defined which combine other getters. E.g. see cwdToApi().
 * 6) avoid defining redundant paths in constants. E.g. create cwdToApi() by combining cwdToExt() and extToApi() instead of defining CWD_TO_API.
 */

use IMATHUZH\Qfq\Core\Exception\Thrower;
use IMATHUZH\Qfq\Core\Store\Config;

class Path
{
    // App
    // This should be manually overwritten (using Path::setCwdToApp()) if the CWD is not the one containing the typo3 index.php.
    private static $cwdToApp = null;

    // Extension
    const APP_TO_EXT = 'typo3conf/ext/qfq';

    // API
    const EXT_TO_API = 'Classes/Api';
    const API_TO_APP = '../../../../../'; // TODO: not relative to ext

    // Icons
    const EXT_TO_GFX_INFO_FILE = 'Resources/Public/icons/note.gif';
    const EXT_TO_PATH_ICONS = 'Resources/Public/icons';

    // Annotate
    const EXT_TO_HIGHLIGHT_JSON_DIR = 'Resources/Public/Json';

    // Twig
    const EXT_TO_TWIG_TEMPLATES = 'Resources/Public/twig_templates';

    // QFQ Project dir
    private static $appToProject = null;
    private const APP_TO_PROJECT_DEFAULT = '../';
    private const APP_TO_PROJECT_MIGRATE = 'fileadmin/protected/qfqProject';
    const PROJECT_TO_FORM = 'form';
    const PROJECT_DIR_TO_REPORT = 'report';

    // Config
    const APP_TO_TYPO3_CONFIG = 'typo3conf';

    // Log files
    private static $cwdToLog  = null; // TODO: does it make sense to have it rel to CWD? to broad?
    private static $overloadAbsoluteQfqLogFile = null;
    private static $overloadAbsoluteMailLogFile = null;
    private static $overloadAbsoluteSqlLogFile = null;
    private const LOG_TO_QFQ_LOG_FILE_DEFAULT = 'qfq.log';
    private const LOG_TO_MAIL_LOG_FILE_DEFAULT = 'mail.log';
    private const LOG_TO_SQL_LOG_FILE_DEFAULT = 'sql.log';
    private const PROJECT_TO_LOG_DEFAULT = 'log';
    private const APP_TO_LOG_DEPRECATED = 'fileadmin/protected/log';

    // Thumbnail
    const APP_TO_SYSTEM_THUMBNAIL_DIR_SECURE_DEFAULT = 'fileadmin/protected/qfqThumbnail';
    const APP_TO_SYSTEM_THUMBNAIL_DIR_PUBLIC_DEFAULT = 'typo3temp/qfqThumbnail';
    const APP_TO_THUMBNAIL_UNKNOWN_TYPE = 'typo3/sysext/frontend/Resources/Public/Icons/FileIcons/';

    // Send Email
    const EXT_TO_SEND_EMAIL_FILE = '/Classes/External/sendEmail';

    /**
     * Manually set the paths which are not constant nor can be inferred by other paths.
     * This function must be called at the beginning of every entry point, to tell the Path class, where things are
     * relative to the CWD.
     *
     * @param string $cwdToApp
     * @throws \CodeException
     * @throws \UserFormException
     */
    public static function setMainPaths(string $cwdToApp)
    {
        self::setCwdToApp($cwdToApp);
        self::findAndSetProjectPath();
        self::findAndSetLogPath();
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function absoluteApp(/* path parts to append */): string
    {
        return self::realpath(self::cwdToApp(func_get_args()));
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function absoluteExt(/* path parts to append */): string
    {
        return self::realpath(self::cwdToExt(func_get_args()));
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function absoluteLog(/* path parts to append */): string
    {
        return self::realpath(self::cwdToLog(func_get_args()));
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function absoluteSqlLogFile(): string
    {
        if (is_null(self::$overloadAbsoluteSqlLogFile)) {
            return self::absoluteLog(self::LOG_TO_SQL_LOG_FILE_DEFAULT);
        }
        return self::$overloadAbsoluteSqlLogFile;
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function absoluteQfqLogFile(): string
    {
        if (is_null(self::$overloadAbsoluteQfqLogFile)) {
            return self::absoluteLog(self::LOG_TO_QFQ_LOG_FILE_DEFAULT);
        }
        return self::$overloadAbsoluteQfqLogFile;
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function absoluteMailLogFile(): string
    {
        if (is_null(self::$overloadAbsoluteMailLogFile)) {
            return self::absoluteLog(self::LOG_TO_MAIL_LOG_FILE_DEFAULT);
        }
        return self::$overloadAbsoluteMailLogFile;
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function cwdToLog(/* path parts to append */): string
    {
        self::enforcePathIsSet(self::$cwdToApp);
        return self::join(self::$cwdToLog, func_get_args());
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function cwdToApp(/* path parts to append */): string
    {
        self::enforcePathIsSet(self::$cwdToApp);
        return self::join(self::$cwdToApp, func_get_args());
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function appToProject(/* path parts to append */): string
    {
        self::enforcePathIsSet(self::$appToProject);
        return self::join(self::$appToProject, func_get_args());
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function cwdToProject(/* path parts to append */): string
    {
        return self::cwdToApp(self::appToProject(func_get_args()));
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function cwdToExt(/* path parts to append */): string
    {
        return self::cwdToApp(self::APP_TO_EXT, func_get_args());
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function extToApi(/* path parts to append */): string
    {
        return self::join(self::EXT_TO_API, func_get_args());
    }

    /**
     * @return string
     * @throws \UserFormException
     */
    public static function cwdToApi(/* path parts to append */): string
    {
        return self::cwdToExt(self::extToApi(func_get_args()));
    }

    /**
     * @param string $newPath
     * @throws \UserFormException
     */
    public static function setAbsoluteSqlLogFile(string $newPath)
    {
        self::enforcePathIsAbsolute($newPath);
        self::$overloadAbsoluteSqlLogFile = $newPath;
    }

    /**
     * @param string $newPath
     * @throws \UserFormException
     */
    public static function setAbsoluteQfqLogFile(string $newPath)
    {
        self::enforcePathIsAbsolute($newPath);
        self::$overloadAbsoluteQfqLogFile = $newPath;
    }

    /**
     * @param string $newPath
     * @throws \UserFormException
     */
    public static function setAbsoluteMailLogFile(string $newPath)
    {
        self::enforcePathIsAbsolute($newPath);
        self::$overloadAbsoluteMailLogFile = $newPath;
    }

    /**
     * Override the paths of sql.log, qfq.log, mail.log using the values from the config file or Typo3.
     *
     * @throws \UserFormException
     */
    public static function overrideLogPathsFromConfig()
    {
        // QFQ log
        $absoluteQfqLogFile = Config::get(SYSTEM_QFQ_LOG_PATHFILENAME);
        if (!empty($absoluteQfqLogFile)) {
            self::setAbsoluteQfqLogFile(self::joinIfNotAbsolute(self::absoluteApp(), $absoluteQfqLogFile));
        }

        // Mail log
        $absoluteMailLogFile = Config::get(SYSTEM_MAIL_LOG_PATHFILENAME);
        if (!empty($absoluteMailLogFile)) {
            self::setAbsoluteMailLogFile(self::joinIfNotAbsolute(self::absoluteApp(), $absoluteMailLogFile));
        }

        // SQL log
        $absoluteSqlLogFile = Config::get(SYSTEM_SQL_LOG_PATHFILENAME);
        if (!empty($absoluteSqlLogFile)) {
            self::setAbsoluteSqlLogFile(self::joinIfNotAbsolute(self::absoluteApp(), $absoluteSqlLogFile));
        }
    }

    /**
     * Return the second path if it is absolute. Otherwise concatenate and return the two paths.
     *
     * @param string $path1
     * @param string $path2
     * @return string
     * @throws \UserFormException
     */
    public static function joinIfNotAbsolute (string $path1, string $path2): string
    {
        if ($path2 !== '' && $path2[0] === '/') {
            return $path2;
        }
        return self::join($path1, $path2);
    }

    /**
     * Join the arguments together as a path.
     * The second argument may be an array of parts. (further arguments are ignored in that case)
     *
     * @return string
     * @throws \UserFormException  Don't remove.
     */
    public static function join(/* path parts to append */): string
    {
        $parts = func_get_args();

        // concatenate all parts (arrays are flattened)
        $path = '';
        array_walk_recursive($parts, function ($part) use (&$path) {

            // filter out empty string and null arguments
            if (is_null($part) || $part === '') {
                return;
            }

            // first part added without '/'
            if ($path === '') {
                $path .= $part;
            } else {

                // trailing path parts may not be absolute
                if ($part[0] === '/') {
                    throw new \UserFormException(json_encode([
                        ERROR_MESSAGE_TO_USER => 'Failed: Join path.',
                        ERROR_MESSAGE_TO_DEVELOPER => "Trailing path parts may not start with '/'. Trailing part: '$part'."]),
                        ERROR_PATH_INVALID);
                }

                $path .= '/' . $part;
            }
        });

        // remove multiple occurrences of '/'
        return preg_replace('/\/{2,}/','/', $path);
    }

    ///////////////////////////////////////////////////  Private  //////////////////////////////////////////////////////

    /**
     * @param string $newPath
     */
    private static function setCwdToApp(string $newPath)
    {
        self::$cwdToApp = $newPath;
    }

    /**
     * @param string $newPath
     */
    private static function setAppToProject(string $newPath)
    {
        self::$appToProject = $newPath;
    }

    /**
     * @param string $newPath
     * @return string
     */
    private static function setCwdToLog(string $newPath)
    {
        self::$cwdToLog = $newPath;
    }

    /**
     * Searches these places for log directory:
     *   1) project-directory/log
     *   2) fileadmin/protected/log
     * If not found create log dir in: project-directory/log
     *
     * @throws \UserFormException
     */
    private static function findAndSetLogPath()
    {
        // search log dir in project
        $cwdToLog = self::cwdToProject(self::PROJECT_TO_LOG_DEFAULT);
        if (file_exists($cwdToLog)) {
            self::setCwdToLog($cwdToLog);

        // search log dir in fileadmin/protected
        } elseif (file_exists(self::cwdToApp(self::APP_TO_LOG_DEPRECATED))) {
            self::setCwdToLog(self::cwdToApp(self::APP_TO_LOG_DEPRECATED));

        // create default log dir
        } else {
            HelperFile::createPathRecursive($cwdToLog);
            self::setCwdToLog($cwdToLog);
        }
    }

    /**
     * Read the project location from qfq.project.path.php or create the file with default path.
     *
     * @throws \CodeException
     * @throws \UserFormException
     */
    private static function findAndSetProjectPath()
    {
        // does qfq.project.path.php exist? => read path
        $cwdToProjectPathFile = self::cwdToApp(PROJECT_PATH_PHP_FILE);
        if (HelperFile::isReadableException($cwdToProjectPathFile)) {
            self::setAppToProject(HelperFile::include($cwdToProjectPathFile));

        // does the deprecated config.qfq.php exist? => migrate to qfq.json
        } elseif (HelperFile::isReadableException(self::cwdToApp(self::APP_TO_TYPO3_CONFIG, CONFIG_QFQ_PHP))) {
            self::setAppToProject(self::APP_TO_PROJECT_MIGRATE);
            self::writeProjectPathPhp();
            Config::migrateConfigPhpToJson();

        // create qfq.project.path.php at default location
        } else {
            self::setAppToProject(self::APP_TO_PROJECT_DEFAULT);
            self::writeProjectPathPhp();
        }
    }

    /**
     * Write the project path configuration file to the project directory.
     *
     * @throws \UserFormException
     */
    private static function writeProjectPathPhp()
    {
        $appToProject = self::appToProject();
        $fileContent = <<<EOF
<?php

/**
QFQ project path configuration

******************************************************-----********************************
* ATTENTION: The files in the project directory should NOT be served by your http server! *
******************************************************-----********************************

Only exception: The app directory inside the project directory may be served.
*/

return '$appToProject'; // path relative to app directory (i.e. location of this file).
EOF;
        HelperFile::file_put_contents(self::cwdToApp(PROJECT_PATH_PHP_FILE), $fileContent);
    }

    /**
     * Throw an exception if the given path is not set (i.e. === null).
     *
     * @param $path
     * @throws \UserFormException
     */
    private static function enforcePathIsSet($path) {
        if(is_null($path)) {
            Thrower::userFormException('Path accessed before set.', 'Make sure Path::setMainPaths() is called before.');
        }
    }

    /**
     * Throw exception if path does not start with '/'
     *
     * @param $path
     * @throws \UserFormException
     */
    private static function enforcePathIsAbsolute(string $path) {
        if($path !== '' && $path[0] === '/') {
            return;
        }
        Thrower::userFormException('Path is not absolute', "Path does not start with '/' : $path");
    }

    /**
     * PHP System function: realpath() with QFQ exception
     *
     * @param string $path
     * @return string
     * @throws \UserFormException
     */
    private static function realpath(string $path): string
    {
        $result = realpath($path);
        if ($result === false) {
            Thrower::userFormException('Path not found.', "Either file/dir not exists or access not permitted: $path.");
        }
        return $result;
    }
}